#!/usr/bin/env python
# ex:ts=4:sw=4:sts=4:et
# -*- tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*-
"""
BitBake 'Fetch' git implementation

Copyright (C) 2005 Richard Purdie

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place, Suite 330, Boston, MA 02111-1307 USA. 
"""

import os, re
import bb
from   bb    import data
from   bb.fetch import Fetch
from   bb.fetch import FetchError

def prunedir(topdir):
    # Delete everything reachable from the directory named in 'topdir'.
    # CAUTION:  This is dangerous!
    for root, dirs, files in os.walk(topdir, topdown=False):
        for name in files:
            os.remove(os.path.join(root, name))
        for name in dirs:
            os.rmdir(os.path.join(root, name))

def runfetchcmd(cmd, d, quiet = False):

    bb.debug(1, "Running %s" % cmd)

    # Need to export PATH as git is likely to be in metadata paths 
    # rather than host provided
    cmd = 'export PATH=%s; %s' % (data.expand('${PATH}', d), cmd)

    # redirect stderr to stdout
    stdout_handle = os.popen(cmd + " 2>&1", "r")
    output = ""

    while 1:
        line = stdout_handle.readline()
        if not line:
            break
        if not quiet:
            print line
        output += line

    status =  stdout_handle.close() or 0
    signal = status >> 8
    exitstatus = status & 0xff

    if signal:
        raise FetchError("Fetch command %s failed with signal %s, output:\n%s" % (cmd, signal, output))
    elif status != 0:
        raise FetchError("Fetch command %s failed with exit code %s, output:\n%s" % (cmd, status, output))

    return output

def contains_ref(tag, d):
    output = runfetchcmd("git log --pretty=oneline -n 1 %s -- 2> /dev/null | wc -l" % tag, d, True)
    return output.split()[0] != "0"

def latest_revision(proto, user, host, path, branch, d):
    """
    Compute the HEAD revision for the url
    """
    if user:
        username = user + '@'
    else:
        username = ""

    output = runfetchcmd("git ls-remote %s://%s%s%s %s" % (proto, username, host, path, branch), d, True)
    return output.split()[0]

def getbranch(parm):
    if 'branch' in parm:
        branch = parm['branch']
    else:
        branch = "master"
    if not branch:
        branch = "master"

    return branch

def gettag(parm, proto, user, host, path, branch, d):
    if 'tag' in parm:
        tag = parm['tag']
    else:
        tag = latest_revision(proto, user, host, path, branch, d)
    if not tag or tag == "master":
        tag = latest_revision(proto, user, host, path, branch, d)

    return tag

def getprotocol(parm):
    if 'protocol' in parm:
        proto = parm['protocol']
    else:
        proto = ""
    if not proto:
        proto = "rsync"

    return proto

def localfile(url, d):
    """Return the filename to cache the checkout in"""
    (type, host, path, user, pswd, parm) = bb.decodeurl(data.expand(url, d))

    #if user sets localpath for file, use it instead.
    if "localpath" in parm:
        return parm["localpath"]

    proto = getprotocol(parm)
    branch = getbranch(parm)
    tag = gettag(parm, proto, user, host, path, branch, d)

    return data.expand('git_%s%s_%s.tar.gz' % (host, path.replace('/', '.'), tag), d)

class Git(Fetch):
    """Class to fetch a module or modules from git repositories"""
    def supports(url, d):
        """Check to see if a given url can be fetched with cvs.
           Expects supplied url in list form, as outputted by bb.decodeurl().
        """
        (type, host, path, user, pswd, parm) = bb.decodeurl(data.expand(url, d))
        return type in ['git']
    supports = staticmethod(supports)

    def localpath(url, d):

        return os.path.join(data.getVar("DL_DIR", d, 1), localfile(url, d))

    localpath = staticmethod(localpath)

    def go(self, d, urls = []):
        """Fetch urls"""
        if not urls:
            urls = self.urls

        for loc in urls:
            (type, host, path, user, pswd, parm) = bb.decodeurl(data.expand(loc, d))

            proto = getprotocol(parm)
            branch = getbranch(parm)
            tag = gettag(parm, proto, user, host, path, branch, d)

            if user:
                username = user + '@'
            else:
                username = ""

            gitsrcname = '%s%s' % (host, path.replace('/', '.'))

            repofilename = 'git_%s.tar.gz' % (gitsrcname)
            repofile = os.path.join(data.getVar("DL_DIR", d, 1), repofilename)
            repodir = os.path.join(data.expand('${GITDIR}', d), gitsrcname)

            coname = '%s' % (tag)
            codir = os.path.join(repodir, coname)

            if not os.path.exists(repodir):
                if Fetch.try_mirror(d, repofilename):    
                    bb.mkdirhier(repodir)
                    os.chdir(repodir)
                    runfetchcmd("tar -xzf %s" % (repofile),d)
                else:
                    runfetchcmd("git clone -n %s://%s%s%s %s" % (proto, username, host, path, repodir),d)

            os.chdir(repodir)
            # Remove all but the .git directory
            if not contains_ref(tag, d):
                runfetchcmd("rm * -Rf", d)
                runfetchcmd("git fetch %s://%s%s%s %s" % (proto, username, host, path, branch), d)
                runfetchcmd("git fetch --tags %s://%s%s%s" % (proto, username, host, path), d)
                runfetchcmd("git prune-packed", d)
                runfetchcmd("git pack-redundant --all | xargs -r rm", d)

            os.chdir(repodir)
            bb.note("Creating tarball of git repository")
            runfetchcmd("tar -czf %s %s" % (repofile, os.path.join(".", ".git", "*") ),d)

            if os.path.exists(codir):
                prunedir(codir)

            bb.mkdirhier(codir)
            os.chdir(repodir)
            copath = os.path.join(codir, "git", "")
            runfetchcmd("git read-tree %s" % (tag),d)
            runfetchcmd("git checkout-index -q -f --prefix=%s -a" % (copath),d)
            bb.mkdirhier(os.path.join(copath, ".git", ""))
            runfetchcmd("LANG=\"en\" git log %s --max-count=1 --date=iso -- > %s.git/last_commit_info" % (tag, copath),d)
            runfetchcmd("echo %s > %s.git/branch" % (branch, copath),d)

            os.chdir(codir)
            bb.note("Creating tarball of git checkout")
            runfetchcmd("tar -czf %s %s" % (self.localpath(loc, d), os.path.join(".", "*") ),d)
